/**
 * Skypager.Helper
 *
 * The Skypager.Helper is a way of specifying the interface and behavior of
 * different categories of javascript modules used by the project.  The different
 * categories of modules such as Actions or Models can provide their own DSLs through
 * the helper system.
*/
import Skypager from '../index'
import * as util from '../util'
import { dirname, extname, resolve, join } from 'path'
import { EventEmitter } from 'fbemitter'

import isObject from 'lodash/isObject'
import get from 'lodash/get'
import result from 'lodash/result'

export default class Helper extends EventEmitter {
  static decorateRegistry (registry) {
    if (this.registryInterface) {
      defineProperties(registry, this.registryInterface(registry))
    }
  }

  /*
  * Creates a Helper from a Definition object that was
  * created by a required' script file from one of the dedicated
  * locations for the type of helper.
  */
  static fromDefinition (uri, definition, options) {
    if(definition && options.required) {
      Object.defineProperty(definition, 'helperExport', {
        enumerable: false,
        get: function() {
          return options.required
        }
      })
    }

    let helper = new this(uri, Object.assign(options, {definition}))

    Object.defineProperties(definition, {
      _helper: {
        enumerable: false,
        get: function(){ return helper }
      },
      _requirePath: {
        enumerable: false,
        get: function(){ return uri }
      }
    })

    return helper
  }

  constructor (uri, options = {}) {
    super()
    let helper = this
    let raw

    util.assign(this, {
      get raw () { return raw },

      set raw (val) {
        raw = val
        //helper.contentDidChange(helper)
      }
    })

    this.hidden('uri', uri)
    this.hidden('options', options)
    this.hidden('project', () => options.project)
    this.getter('owner', () => options.owner)
    this.getter('required', (() => options.required || this.require(uri)), true)

    let definition = options.definition

    // before returning the definition make sure to call any configure methods we have
    this.lazy('definition', function () {
      let d = options.definition || this.required.definition

      if (d && d.configure) { d.configure() }

      if(this.required && this.required.config){
        Object.keys(this.required.config).forEach(key => {
          if(d && d.config && d.config[key]){
            d.config[key] = Object.assign(this.required.config[key], d.config[key])
          } else {
            d.config[key] = this.required.config[key]
          }
        })
      }

      return d
    })

    this.id = this.paths.relative.replace(this.extension, '')


    let _this = this

    Object.defineProperty(_this, 'api', {
      configurable: true,
      enumerable: false,
      get: function() {
        return this.buildAPI(options.api, this.required) || (function() {  })
      }
    })
  }

  /**
   * creates a subscription to an event
   *
   * @param String event name of the event to monitor
   * @Param function callback callback to execute when event occurs
   */
  on(event, callback) {
    return this.addListener(event, callback)
  }

  //TODO: implement
  //see https://github.com/facebook/emitter#__emittosubscriptionsubscription-eventtype-args
  //__emitToSubscription(subscription, eventType, ...args) {}
  //
  get(...args) {
    return get(this, ...args)
  }

  result(...args) {
    return result(this, ...args)
  }
  /**
  * Every helper should expose an api with a function which is responsible
  * for handling calls to the run function that get dispatched to the helper.
  *
  */
  run (...args) {
    let context = args[ args.length - 1]

    if(process.env.NODE_ENV !== 'test' && !(isObject(context) && context.__registryPath)){
      console.warn(
        `Calling helpers directly is not supported. Use the registry run interface. From ${ this.id }`
      )
    }

    let project = context.project || (args[0] && typeof(args[0]) === 'object' && args[0].project) || this.project
    let fn = project ? this.runner.bind(project) : this.runner

    return util.noConflict(fn, {
      project,
      Skypager,
      skypager: Skypager,
      util,
      currentHelper: this
    })(...args)
  }

  get idPath () {
    return this.id.replace('-', '_').replace('/', '.')
  }

  hidden (property, value) {
    let key = typeof value === 'function'
      ? 'get'
      : 'value'

    Object.defineProperty(this, property, {
      enumerable: false,
      configurable: false,
      [key]: value
    })
  }

  lazy (...args) { return util.lazy(this, ...args) }

  /**
   * Build the run API for this helper.
   *
   * This is how the helper gets "used" in the project. Each
   * type of helper (e.g. actions) might have its own pipeline
   * of functions to run.  Some standard, some custom and supplied
   * by the helper. This is how we standardize that call stack and
   * ensure that predictable context can be passed with each calls
   * and its non-predictable parameters
   *
   */
  buildAPI (api) {

    if (this.project) {
      this.project.debug(`Building helper API for ${ this.id }`, api)
    }

    if (api) { return api }

    let mod = this.required

    let runner

    // if this helper module exported a function
    if ( typeof mod === 'function' ) {
      runner = this.project ? mod.bind(this.project) : mod

      return Object.assign(runner, {
        runner,
        id: this.id,
        definition: this.definition
      })
    }

    if (typeof mod === 'object' && typeof mod.default === 'function') {
      return Object.assign(mod, {
        runner: this.project ? mod.default.bind(this.project) : mod.default,
        id: this.id,
        definition: this.definition
      })
    }

    if (typeof mod === 'object') {
      return Object.assign(this.definition.api || {}, mod)
    }

    throw ('There was a problem building an API for the helper id: ' + this.id + ' at ' + this.uri)
  }

  get runner () {
    return this.api.runner
  }

  require (uri) {
    let result = module.require(uri)
    return result
  }

  get paths () {
    let asset = this
    return {
      get summary () {
        let parts = asset.uri.split(/\/|\\/)
        return parts.reverse().slice(0,3).reverse().join('/')
      },
      get relative () {
        return asset.uri.replace(asset.owner.root + '/', '')
      },
      get absolute () {
        return resolve(asset.uri)
      }
    }
  }

  contentWillChange (oldContent, newContent) {

  }

  contentDidChange (asset) {

  }

  get extension () {
    return extname(this.uri)
  }

  getter (name, obj, configurable) {
    util.hide.getter(this, name, obj, configurable)
  }
}

Helper.apiMethods = [ ]

const { defineProperties } = Object
